#include <stdint.h>
#include <stdio.h>
#include <propeller2.h>
#define JD_FORMAT 2
#include "tjpgd.c"

typedef uint8_t pixel_t;
typedef struct __using("p2videodrv.spin2") vdrv_t;

#define JDEC_WORKBUF_SIZE (10*1024)

uint16_t origin_x,origin_y,scale;

vdrv_t vdrv;
uint8_t display[vdrv.DISPLAY_SIZE];
uint8_t region[vdrv.REGION_SIZE];
pixel_t lineBuffers[640*2];
pixel_t screenBuffer[640*480];
uint8_t workBuf[JDEC_WORKBUF_SIZE]; // may need to be large with optimized huffmann
uint8_t vcog;

size_t in_func (    /* Returns number of bytes read (zero on error) */
    JDEC* jd,       /* Decompression object */
    uint8_t* buff,  /* Pointer to the read buffer (null to remove data) */
    size_t nbyte    /* Number of bytes to read/remove */
) {
    if (buff) memcpy(buff,jd->device,nbyte);
    jd->device = ((uint8_t*)jd->device)+nbyte;
    return nbyte;
}

int out_func (
  JDEC* jdec,    /* Pointer to the decompression object */
  void* bitmap,  /* Bitmap to be output */
  JRECT* rect    /* Rectangle to output */
) {
    pixel_t *bmp = bitmap;
    pixel_t *out = screenBuffer + (rect->left+origin_x+(rect->top+origin_y)*640);
    unsigned width = rect->right - rect->left + 1;
    unsigned height = rect->bottom - rect->top + 1;

    do {
        pixel_t *line = out;
        uint16_t x=width;
        do {
            *line = *bmp;
            *line++;
            *bmp++;
        } while (--x);
        out += 640;
    } while (--height);
    return 1;
}

uint32_t do_decomp(uint8_t *file) {
    uint32_t time = _cnt();
    JDEC jdec = {};
    jd_prepare(&jdec,&in_func,workBuf,JDEC_WORKBUF_SIZE,file);
    jd_decomp(&jdec,&out_func,scale);
    return _cnt() - time;
}

void main() {

    vcog = vdrv.initDisplay( -1,           // the cogid to use (-1 = auto-allocate)
                            display,       // the display structure address in HUB RAM
                            vdrv.VGA,       // video output type (VGA/DVI etc)
                            32,  // base pin number (hsync pin) of VGA pin group
                            32+4,    // VSYNC pin (not used for DVI)
                            0,              // display flags
                            lineBuffers, // address of the consecutive two scan line buffers in HUB RAM
                            640*sizeof(pixel_t),     // size of a single scan line buffer in bytes
                            vdrv.getTiming(vdrv.RES_640x480), // obtain stock timing to use, (or create custom timing instead)
                            0, // optional external memory mailbox address in HUB RAM (0=none)
                            vdrv.initRegion( // setup a single gfx region as the display list
                                        region,        // region structure address in HUB RAM
                                        vdrv.LUMA_WHITE,       //
                                        0,              // size of region in scan lines (0=to end of screen)
                                        0, // region specific flags (if enabled text flashes if BG colour > 7)
                                        0,       // address of default palette to be used by region
                                        0,              // address of default font to be used by this region
                                        8,              // number of scan lines in font
                                        screenBuffer,    // address of screen buffer in HUB RAM
                                        0)              // link to next region, NULL = last region
                            );
    for (;;) {
        unsigned total = 0;
        #define DOTEST(file) {unsigned onetime = do_decomp(&file);total+=onetime;printf(#file " decompd in %u cycles!\n",onetime);}
        origin_x = 0,origin_y = 0,scale = 0;
        DOTEST(reimu_jpg);
        DOTEST(colorburst_jpg);
        DOTEST(lighthouse_jpg);
        origin_x = 0,origin_y = 0,scale = 0;
        DOTEST(edge_jpg);
        origin_x = 320,origin_y = 0,scale = 1;
        DOTEST(reimu_jpg);
        origin_x = 0,origin_y = 240,scale = 1;
        DOTEST(colorburst_jpg);
        origin_x = 320,origin_y = 240,scale = 1;
        DOTEST(lighthouse_jpg);

        printf("TOTAL test suite: %u cycles!!!!\n",total);
    }
}

__pasm {
reimu_jpg file "reimu480p.jpg"
colorburst_jpg file "colorburst480p.jpg"
lighthouse_jpg file "lighthouse480p.jpg"
edge_jpg file "edge240p.jpg"
}
